/*
 * John William Lynch
 * CSC 480 : Assignment 2
 * BoardNode Class that takes in a Board object with it's current stae
 */

package Assignment_2;

//Array list to maintain all of the children associated with the requested parent node.
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;

import Assignment1_NodeDataStructure.A1SearchComparator;
import Assignment1_NodeDataStructure.FullNode;

public class BoardNode {

	/*
	 * This classes primary purpose is to act as a helper to the
	 * AI in order for the successor and recursive MiniMax methods to work effectively.
	 * 
	 * The items it requires to run properly are both a parentNode and a childNode List
	 * so that it can effectively iterate through those objects during the algorithm
	 * of MiniMax and determine the most effective move.
	 * 
	 */
	
	//I added these in later to help with the getSucessors function, what they do is keep track
	//of which x and y variables were changed in the getSuccessors function and will be on stand-by
	//until a winning node from the group decided by the heuristic of each one, and once that
	//winner is decided it's x and y values will be taken.
	
	//Parent node of the BoardNode
	BoardNode parentNode;

	//The state of the currentBoard[][] ie: what the 2D array currently looks like in numerical fashion:
	Board currentBoard;
	
	//Two extra coordinate variables that are set at the time a value is read in as horizontal / vertical
	//in the get successors function
	int xCoordChanged;
	int yCoordChanged;
	
	//Not all Nodes have this heuristic (like the very starting one), but after we create the firstNode,
	//all node should try and see how to compare to each other based on the player score - ai score
	int currentNodeHeuristic;
	
	//An array list of children that each node has
	Comparator<BoardNode> compare = new BoardComparator();
	PriorityQueue <BoardNode> children = new PriorityQueue<BoardNode>(500, compare);
	
	//Constructor object of the BoardNode Class : it will simply take in the Board object, as there
	//is lots of functionality attached to that object and class. It took a bit more time and space to figure out
	//how to do this by just using the 2D Array of ints[], so I found this to be more appropriate.
	public BoardNode(Board currentBoard) {
		this.currentBoard = currentBoard;
	}
	
	//Add children to the list
	public void addChild(BoardNode child) {
		this.children.add(child);
	}
	//Method to set the parent of this node:
	public void setParent (BoardNode parent) {
		this.parentNode = parent;
	}
	
	//Method to get the parent of the node:
	public BoardNode getParent () {
		return this.parentNode;
	}
	
	//Method to set the current board game:
	public void setBoard(Board game) {
		this.currentBoard = game;
	}
	//Methods to set and get the heuristics for our comparator class
	public void setNodeHeuristic(int score) {
		this.currentNodeHeuristic = score;
	}
	
	public int getNodeHeuristic() {
		return this.currentNodeHeuristic;
	}
	
	/*
	 * The get successors function which is useful in our MiniMax.java class
	 * This function will return a List<BoardNodes> which will allow which-ever
	 * Node called it, and in its specific state as Board is passed to these objects
	 * to determine the proper way to proceed
	 * The two values it takes in again are a startNode and boolean value; playerTurn,
	 * As this boolean value tells the function what values it can pass off (true or false)
	 * to determine whose turn it is
	 * 
	 * Remember that our constructor for the Board.java class consists of the following:
	 * public Board (int[][] currentBoard, int rows, int columns, boolean playerTurn, int scoreForPlayer, int scoreForAI) {
	 * 
	 * Therefore, we need the startNode to parse out its relevant Board object, so we can also take from that attribute the
	 * state of the board, the rows, columns, whose turn it is, and the respective scores for each player
	 */
	public void getSuccessors(){
		
		//Create default placeholder for all node children : I originally used a list, but I decided to use
		//a priority queue with my custom comparator class so that I can 
		
		
		//To implement this a new comparator was created and will be used (based on h1 code: the path farthest away from the goal state)

		//Reduce the Node given to us into a board that contains variables we need for later on
		//The most important one of these variables will be the int[][] variable representing the current game state
		Board newBoard = this.currentBoard;
		
		//What else do we need to pass over to the constructor: extract the state from the board
		int[][] currentBoardState = newBoard.getCurrentState();
		
		//Extract the rows and columns as well, as we need them for the constructor and the below for loop
		int rows = newBoard.rows;
		int columns = newBoard.columns;
		
		//Also grab the previous scores of the ai and player
		int playerScore = newBoard.scoreForPlayer;
		int aiScore = newBoard.scoreForAI;
		
		//We now need to do a large for loop that will iterate over the positions of our board using the similar
		//row and column for loop. During this iteration, we need to check for any location where we can add 
		//either a HORIZONTAL LINE or a VERTICAL LINE. We will try and see if we can add in a horizontal line first
		for (int row = 0; row < rows; row++) {
			
			for (int column = 0; column < columns; column++) {
				
				/*
				 * Using previously defined logic, we need to double check what we remember about the rows and columns
				 * First of all, given a basic 1x1 
				 * 
				 *  + * + row 0
				 *  * 1 * row 1
				 *  + * + row 2
				 *  0 1 2
				 *  
				 *  We see that at row0, row == even, the only location we can enter is at column 1, column == odd
				 *  This gives us a HORIZONTAL line at 0,1, and 2,1
				 *  This means, that we can only do a horizontal line when our row is even, and our column is odd
				 *  
				 *  For our column, we can only enter in a vertical line at row 1, row == odd, and column 0 and column 2, column == even
				 *  This means, that we can only do a vertical line when our column is even, and our row is odd
				 *  
				 *  As this is repeated for all values within our game board just with larger rows and columns, we will code this
				 *  logic directly into our for loop and have the AI randomly add in vertical / horizontal lines where it deems fit
				 *  
				 *  We also need to check if the item there is available, meaning that the value at those locations must be equal to 0
				 *  which is what I chose to be my blank value
				 *  
				 *  The time of the AI may change based on which one it adds first, but as we can get an instant horizontal line everytime 
				 *  if we do rows first, I will pick going with horizontal lines first
				 */
				
				//Check if spot is open, and then if row is even, and column is odd : add horizontal line
				if (currentBoardState[row][column] == 0 && (row % 2 == 0 && column % 2 != 0)) {
					
					//Now we will do a successor function, similar to previous assignment, where we create a new node and update values
					
					//Change the array to now have a horizontal line:
					/*
					 * My AI Was automatically finishing my problem without any more prompts, this because it wasn't copying
					 * but setting the new values
					 */
					//Copy over the item to a new int array
					
					int[][] childBoardState = copyOverArray(currentBoardState, rows, columns);
					childBoardState[row][column] = -1;
					
					//Create a NEW board child with the found values
					Board childBoard = new Board(childBoardState, rows, columns, false, playerScore, aiScore);
					
					
					BoardNode childNode = new BoardNode(childBoard);
					
					//Set the x and y coord it changed in this move
					childNode.xCoordChanged = row;
					childNode.yCoordChanged = column;
					
					//Set the val of the heuristic for comparisons
					childNode.setNodeHeuristic(childNode.currentBoard.scoreForPlayer - childNode.currentBoard.scoreForAI);
					
					//Set the parent
					childNode.setParent(this);
					
					//Add into the array list
					this.addChild(childNode);
				}
				
				//Now we do something very similar, but this time it will be for vertical lines
				else if (currentBoardState[row][column] == 0 && (row % 2 != 0 && column % 2 == 0)) {
					//Now we will do a successor function, similar to previous assignment, where we create a new node and update values
					
					//Change the array to now have a horizontal line:
					int[][] childBoardState = copyOverArray(currentBoardState, rows, columns);
					childBoardState[row][column] = -2;
					
					//Create a NEW board child with the found values
					Board childBoard = new Board(currentBoardState, rows, columns, true, playerScore, aiScore);
					
					BoardNode childNode = new BoardNode(childBoard);
					
					//Set the correct x and y of what this childNode found
					childNode.setNodeHeuristic(childNode.currentBoard.scoreForPlayer - childNode.currentBoard.scoreForAI);
					
					//Set the parent
					childNode.setParent(this);
					
					//Add into the array list
					this.addChild(childNode);
					
					//Set the x and y coord it changed in this move
					childNode.xCoordChanged = row;
					childNode.yCoordChanged = column;
					
				
				}
			}
		}
		
	}
	
	//Please note, I had this function inside of my NodeBoard.java class but for some reason whenever
	//I passed it via the nodes it would make an empty amount of parents, so I attempted to use it here
	
	//Helper method for the get successor function that creates a cloned 2d array
		//to be used by the children class : pass over old array with matching row and columns for that array
		public int[][] copyOverArray(int[][] old, int rows, int columns){
			int[][] newArray = new int[rows][columns];
			for (int i = 0; i < rows; i++) {
				for (int j = 0; j < columns; j++) {
					newArray[i][j] = old[i][j];
				}
			}
			return newArray;
		}

}
