/*
 * John William Lynch
 * CSC 480 : Assignment 2
 * MiniMax Algorithm.java file to work with the successor and various other functionality
 * 
 */
package Assignment_2;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;

public class MiniMax {

	 /*
	 * The MiniMax algorithm with Alpha-Beta pruning consists of two parts for execution:
	 * The first part is developing all possible new int[][] states for the AI to view, because
	 * the int[][] currentBoard allows us to dynamically give the board back and forth between
	 * node states and for the algorithm to determine the most effective board based on whose turn it is
	 */
	
	//Two integer values that represent the Min and Max integer values
	/*
	 * I attempted to use my own values for alpha and beta, setting alpha at 9999
	 * and beta at 0, however I felt it was better for testing purposes to instead use the
	 * Integer.MAX and Intger.Min associated values because they seem to do a better job at
	 * what the alpha and beta are supposed to represent than arbitrarily created numbers
	 */
	
	//Integer values to represent the ply (which was given in Main.java by the user)
	int ply;
	
	//Get the current Board (which includes the state[][] values and other important values, such as whose turn it is)
	Board currentGame;
	
	//Create the constructor of the MiniMax class:
	public MiniMax(Board currentGame, int ply) {
		this.currentGame = currentGame;
		this.ply = ply;
	}
	
	
    /*
     * This is my MiniMax SEARCH function that uses a mixture of the heuristic and alpha beta pruning
     * The alpha pruning occurs when the beta <= alpha for MAX, and vice versa for MIN
     * I pass a boolean MAX that tells the AI whose turn it is, and that is swapped
     * in every recursive iteration
     */
	private static int searchMiniMax(BoardNode startNode, int ply, int alpha, int beta, int depth, boolean max)
	{
		
		startNode.getSuccessors();
		
		if(depth > ply || startNode.children.isEmpty())
		{
			startNode.getNodeHeuristic();
			return startNode.currentNodeHeuristic;
		}
		
		//if it is not MAX IE MIN's turn
		if(!max)
		{
			//Identical concept as described below, but with different values as this is MIN
			int bestVal = beta;
			for(BoardNode n : startNode.children)
			{
				int value = searchMiniMax(n, ply, alpha, beta, depth+1, true);
				//If the found value recursively is better than what we determined as the bestVal
				//Then that new one is the new bestVal
				if(value < bestVal)
					bestVal = value;
				//If the best val is also less than the beta, then we know that should be the new beta
				if(bestVal < beta)
					beta = bestVal;
				
				//Using the logic described via the LECTURES, if we see a tree node that is not
				//beneficial (ie: beta less than alpha) we know that we do not want to go down that path 
				//as it will only cause more overhead. 
				if(beta <= alpha)
					
					break;
			}
			startNode.currentNodeHeuristic = bestVal;
			return bestVal;
		}
		else
		{
			int bestVal = alpha;
			for(BoardNode n : startNode.children)
			{
				int value = searchMiniMax(n, ply, alpha, beta, depth+1, false);
				//If the found value recursively is better than what we determined as the bestVal
				//Then that new one is the new bestVal
				if(value > bestVal)
					bestVal = value;
				//If the newly found bestVal is also greater than what we determined to be max, then also remove it
				if(bestVal > alpha)
					alpha = bestVal;
				//At this step we do pruning, as if the beta is less than the alpha no need to
				//go further into that tree
				if(beta <= alpha)
					
					break;
			}
			//Set the Heuristic (the score - ai) to the new heuristic so the node's can compare with each other
			startNode.currentNodeHeuristic = bestVal;
			return bestVal;
		}
	}
	
	/*
	 * This run function takes in a BoardNode as it's main value alongside the amount of determine
	 * PLY that the user stated at the start. It then passes the values over to the
	 * searching algorithm of MiniMax, and I have a few notes described in the video regarding my ALG,
	 * but for some reason it seems like the AI I designed would only increment to forward or behind based on changes
	 * I made here. If I didn't return the children coordinates, the AI would use the largest coordinates to find itself and go from there
	 * If I used the 
	 */
	public static int[] run(BoardNode givenNode, int ply)
	{
		
		BoardNode startNode = new BoardNode(givenNode.currentBoard);
		
		int determinedBestVal = searchMiniMax(startNode, ply, Integer.MIN_VALUE, Integer.MAX_VALUE, 0, true);
		
		//Store the decision of the AI inside of a two int large array that maps to an x and y
		int aiDecision[] = new int[2];
		
		startNode.getSuccessors();
		
		for(BoardNode n : startNode.children)
		{
			if(n.getNodeHeuristic() == determinedBestVal)
			{
				//If the heuristic is the best possible heursitic we can achieve
				//then add that into our decision and return it
				aiDecision[0] = n.xCoordChanged;
				aiDecision[1] = n.yCoordChanged;
				return aiDecision;
			}
		}
		return aiDecision;
	}
	


}
